import warnings
from copy import deepcopy
from typing import Union, Optional, Iterable, Tuple

import numpy as np
from numpy import typing as npt


class Model:

    def __init__(self, data: Union[dict, "Model"]):
        if isinstance(data, Model):
            # copy constructor
            for k, v in data.__dict__.items():
                setattr(self, k, deepcopy(v))

        else:
            # from dict
            self.r = np.array(data['r'])
            self.F = np.array(data['F'])
            self.Q = np.array(data['Q'])
            assert self.Q.ndim == 1, "Q must be a one-dimensional array"

            self._std = None
            self._Di = None
            self._D = None
            if 'Di' in data:
                assert 'D' not in data, "Di and D cannot be both in model data"
                self.Di = np.array(data['Di'])
            else:
                self.D = np.array(data['D'])

    @property
    def std(self):
        if self._std is None:
            self._std = np.sqrt(self.Q + np.diag(self.F @ self.D @ self.F.transpose()))
        return self._std

    @property
    def D(self):
        if self._D is None:
            # calculate inverse first
            D = np.linalg.inv(self.Di)
            D = (D + D.T)/2
            self._D = D
        return self._D

    @D.setter
    def D(self, value: npt.NDArray):
        self._D = value
        self._Di = None
        self._std = None

    @property
    def Di(self):
        if self._Di is None:
            # calculate inverse first
            Di = np.linalg.inv(self.D)
            Di = (Di + Di.T)/2
            self._Di = Di
        return self._Di

    @Di.setter
    def Di(self, value: npt.NDArray):
        self._Di = value
        self._D = None
        self._std = None

    def to_dict(self, fields: Optional[Iterable]=None,
                as_list: bool=False,
                normalize=False) -> dict:
        # normalize
        alpha = np.max(self.std) ** 2 if normalize else 1.0
        if fields:
            d = {f: getattr(self, f) for f in fields}
            if normalize:
                for f in ['Q', 'D']:
                    if f in fields:
                        d[f] /= alpha
        else:
            d = { 'r': self.r, 'D': self.D / alpha, 'F': self.F, 'Q': self.Q / alpha, 'std': self.std }
        return {k: v.tolist() for k, v in d.items()} if as_list else d

    def unconstrained_frontier(self, x_bar: float=1.):
        q = np.diag(self.Q) + self.F @ self.D @ self.F.transpose()
        b = np.vstack((self.r, np.ones((len(self.r))))).transpose()
        try:
            bsb = b.transpose() @ np.linalg.solve(q, b)
            bsb_inv = np.linalg.inv(bsb)
            a = bsb_inv[0, 0]
            b = -bsb_inv[0, 1]
            c = bsb_inv[1, 1]
            mu_star = b * x_bar / a
            sigma_0 = np.sqrt(c - b ** 2 / a) * x_bar
        except np.linalg.LinAlgError:
            zero_q = self.Q == 0.0
            zero_f = np.sum(self.F, axis=1) == 0
            if np.all(zero_f == zero_q):
                # zero return and zero variance
                reduced_model = Model({'r': self.r[~zero_q], 'Q': self.Q[~zero_q],
                                       'F': self.F[~zero_q,:], 'D': self.D })
                return reduced_model.unconstrained_frontier(x_bar)
            else:
                raise ValueError("Risk-free asset is not supported in this version")
        return a, mu_star, sigma_0

    def return_and_variance(self, x: npt.NDArray) -> Tuple[float, float]:
        # normalize for calculating return and standard deviation
        value = sum(x)
        if value < 0:
            value = -value
            warnings.warn("Total portfolio is negative")
        elif value == 0.:
            value = 1
            warnings.warn("Total portfolio is zero")
        mu = np.dot(x, self.r) / value
        v = self.F.transpose() @ x
        std = np.sqrt(np.dot(self.Q * x, x) + np.dot(self.D @ v, v)) / value
        return mu, std



import unittest
import numpy as np


class TestModel(unittest.TestCase):

    def test_constructor_1(self):

        from pyoptimum.model import Model

        data = {
            'Q': np.random.normal(size=(5,)),
            'F': np.random.normal(size=(5,3)),
            'D': np.random.normal(size=(3,3)),
            'r': np.random.normal(size=(5,))
        }
        data['Q'] = data['Q'] ** 2
        data['D'] = data['D'].T @ data['D']

        # data does not have Di
        model = Model(data)
        self.assertIsNone(model._Di)

        # will calculate Di
        di = model.Di
        self.assertIsNotNone(model._Di)
        np.testing.assert_array_almost_equal(model.D @ model.Di, np.eye(3))

        # make sure it is cached
        di2 = model.Di
        self.assertIs(di, di2)

        # will set D
        D = np.random.normal(size=(3, 3))
        D = D.T @ D
        model.D = D
        self.assertIsNone(model._Di)

        # will calculate Di
        di = model.Di
        self.assertIsNotNone(model._Di)
        np.testing.assert_array_almost_equal(model.D @ model.Di, np.eye(3))

        # make sure it is cached
        di2 = model.Di
        self.assertIs(di, di2)

        # will set Di
        Di = np.random.normal(size=(3, 3))
        Di = Di.T @ Di
        model.Di = Di
        self.assertIsNone(model._D)

        # will calculate D
        d = model.D
        self.assertIsNotNone(model._D)
        np.testing.assert_array_almost_equal(model.D @ model.Di, np.eye(3))

        # make sure it is cached
        d2 = model.D
        self.assertIs(d, d2)

        data = {
            'Q': np.random.normal(size=(5,)),
            'F': np.random.normal(size=(5,3)),
            'Di': np.random.normal(size=(3,3)),
            'r': np.random.normal(size=(5,))
        }
        data['Q'] = data['Q'] ** 2
        data['Di'] = data['Di'].T @ data['Di']

        # data does not have D
        model = Model(data)
        self.assertIsNone(model._D)

        # will calculate D
        d = model.D
        self.assertIsNotNone(model._D)
        np.testing.assert_array_almost_equal(model.D @ model.Di, np.eye(3))

        # make sure it is cached
        d2 = model.D
        self.assertIs(d, d2)

        # test std
        self.assertIsNone(model._std)
        s = model.std
        self.assertIsNotNone(model._std)

        # make sure it is cached
        s2 = model.std
        self.assertIs(s, s2)

        data = {
            'Q': np.random.normal(size=(5,)),
            'F': np.random.normal(size=(5,3)),
            'Di': np.random.normal(size=(3,3)),
            'r': np.random.normal(size=(5,))
        }
        data['Q'] = data['Q'] ** 2
        data['Di'] = data['Di'].T @ data['Di']

        # data does not have D
        model = Model(data)
        self.assertIsNone(model._D)

        # test std
        self.assertIsNone(model._std)
        s = model.std
        self.assertIsNotNone(model._std)

        # make sure it is cached
        s2 = model.std
        self.assertIs(s, s2)

        # a D has been calculated
        self.assertIsNotNone(model._D)

        data = {
            'Q': np.random.normal(size=(5,)),
            'F': np.random.normal(size=(5,3)),
            'D': np.random.normal(size=(3,3)),
            'Di': np.random.normal(size=(3,3)),
            'r': np.random.normal(size=(5,))
        }
        with self.assertRaises(AssertionError):
            Model(data)

    def test_constructor_2(self):

        from pyoptimum.model import Model

        data = {
            'Q': np.random.normal(size=(5,)),
            'F': np.random.normal(size=(5,3)),
            'D': np.random.normal(size=(3,3)),
            'r': np.random.normal(size=(5,))
        }
        data['Q'] = data['Q'] ** 2
        data['D'] = data['D'].T @ data['D']

        # create model
        model_1 = Model(data)

        # copy constructor
        model_2 = Model(model_1)
        np.testing.assert_array_equal(model_1.r, model_2.r)
        np.testing.assert_array_equal(model_1.Q, model_2.Q)
        np.testing.assert_array_equal(model_1.F, model_2.F)
        np.testing.assert_array_equal(model_1.D, model_2.D)
        np.testing.assert_array_equal(model_1.Di, model_2.Di)
        np.testing.assert_array_equal(model_1.std, model_2.std)

        self.assertIsNot(model_1.r, model_2.r)
        self.assertIsNot(model_1.Q, model_2.Q)
        self.assertIsNot(model_1.F, model_2.F)
        self.assertIsNot(model_1.D, model_2.D)
        self.assertIsNot(model_1.Di, model_2.Di)
        self.assertIsNot(model_1.std, model_2.std)

    def test_unconstrained_frontier_and_return(self):

        from pyoptimum.model import Model

        rng = np.random.default_rng(12345)
        data = {
            'Q': rng.normal(size=(5,)),
            'F': rng.normal(size=(5,3)),
            'D': rng.normal(size=(3,3)),
            'r': rng.normal(size=(5,))
        }
        data['Q'] = data['Q'] ** 2
        data['D'] = data['D'].T @ data['D']
        x = rng.random(size=(5,))

        # create model
        model = Model(data)

        # variance and return
        mu, std = model.return_and_variance(x)
        np.testing.assert_array_almost_equal([mu, std], [0.5121689723772288, 2.9559252230222635])

        # unconstrained frontier
        vals = model.unconstrained_frontier()
        np.testing.assert_array_almost_equal(vals, [0.14766907361100318, -0.8407610288104532, 0.5613114301321959])

        # zero return model
        data = {
            'r': np.hstack(([0], model.r, [0])),
            'Q': np.hstack(([0], model.Q, [0])),
            'F': np.vstack((np.zeros((1, 3)), model.F, np.zeros((1, 3)))),
            'D': model.D,
        }
        x = np.hstack(([0], x, [0]))

        # create model
        model = Model(data)

        # variance and return
        mu, std = model.return_and_variance(x)
        np.testing.assert_array_almost_equal([mu, std], [0.5121689723772288, 2.9559252230222635])

        # unconstrained frontier
        vals = model.unconstrained_frontier()
        np.testing.assert_array_almost_equal(vals, [0.14766907361100318, -0.8407610288104532, 0.5613114301321959])

        # holding a zero-return stock
        x[0] = 0.3
        mu, std = model.return_and_variance(x)
        np.testing.assert_array_almost_equal([mu, std], [0.4744052318205996, 2.7379760709896956])

        x[1] = -10.3
        with self.assertWarns(Warning) as w:
            mu, std = model.return_and_variance(x)
        self.assertEqual(str(w.warnings[0].message), "Total portfolio is negative")
        np.testing.assert_array_almost_equal([mu, std], [-3.8480800555552026, 3.726822456664636])

        with self.assertWarns(Warning) as w:
            mu, std = model.return_and_variance(0 * x)
        self.assertEqual(str(w.warnings[0].message), "Total portfolio is zero")
        np.testing.assert_array_almost_equal([mu, std], [0, 0])



import unittest
import os

import requests
import math

from src import pyoptimum

username = 'demo@optimize.vicbee.net'
password = 'optimize'
base_url = os.getenv('TEST_BASE_URL', 'https://optimize.vicbee.net')

class TestBasic(unittest.TestCase):

    def test_constructor(self):

        client = pyoptimum.Client(username=username, password=password)
        self.assertIsNone(client.token)

        client = pyoptimum.Client(token='token')
        self.assertIsNone(client.username)
        self.assertIsNone(client.password)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.Client, username=username)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.Client, username=password)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.Client, token='')

    def test_urls(self):

        answer = 'a/b/c'
        result = pyoptimum.Client.url_join('a', 'b', 'c')
        self.assertEqual(result, answer)

        answer = 'a/b/c'
        result = pyoptimum.Client.url_join('a', '', 'b', '', 'c', '')
        self.assertEqual(result, answer)

        answer = 'a/b/c'
        result = pyoptimum.Client.url_join('a/', '', '/b', '', 'c/', '')
        self.assertEqual(result, answer)

        answer = 'a/b/c'
        result = pyoptimum.Client.url_join('//a//', '', '//b', '', 'c//', '')
        self.assertEqual(result, answer)

        answer = 'a:5000/b/c'
        result = pyoptimum.Client.url_join('//a:5000', '', '//b', '', 'c//', '')
        self.assertEqual(result, answer)

        # remove trailing slashes
        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url + '////')
        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

        # remove trailing slashes
        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url + '////', api='optimize', prefix='api')
        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

        # remove trailing slashes but not in the middle
        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url + '////', api='optimize/api', prefix='')
        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

        # wrong url that is not json
        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url, api='optimizeXY', prefix='api')
        self.assertIsNone(client.token)
        with self.assertRaises(requests.exceptions.HTTPError):
            client.get_token()

    def test_optimize(self):

        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url)

        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

        # wrong password
        client = pyoptimum.Client(username=username, password=password + 'wrong')

        self.assertIsNone(client.token)
        self.assertRaises(requests.exceptions.HTTPError, client.get_token)

        # wrong constructor
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.Client)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.Client,
                          username=username)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.Client,
                          password=password)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.Client,
                          token='')

    def test_models(self):

        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url, api='models')

        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

    def test_auth(self):

        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url, api='models',
                                  auth_url='https://optimize.vicbee.net/auth/api')

        self.assertIsNone(client.token)
        client.get_token()
        self.assertIsNotNone(client.token)

    def test_portfolio(self):

        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url)

        s1 = 0.06
        s2 = 0.03
        rho = 1
        data = {
            'Q': [[s1 ** 2, s1 * s2 * rho], [s1 * s2 * rho, s2 ** 2]],
            'cashflow': 1,
            'mu': 0.11,
            'r': [.14, .08],
        }
        response = client.call('portfolio', data)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        status = response.get('status')
        self.assertEqual(status, 'optimal')

        x = response.get('x')
        self.assertAlmostEqual(x[0], .5)
        self.assertAlmostEqual(x[1], .5)

        self.assertIsNone(client.detail)

        # call with slash
        response = client.call('/portfolio', data)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        # call with errors
        data['r'] = [.14, .08, 0]
        self.assertRaises(pyoptimum.PyOptimumException, client.call, 'portfolio', data)
        self.assertIn('must be an array', client.detail)

    def test_forbidden(self):

        client = pyoptimum.Client(username=username, password=password,
                                  base_url=base_url)

        data = {
            'A': [[2]],
            'blo': [0],
            'bup': [4],
            'c': [-1],
            'xlo': [0],
            'xup': [2]
        }
        self.assertRaises(requests.exceptions.HTTPError, client.call, 'lp', data)
